# ------------------------------------------------------------------
# Licensed under the ISC License. See LICENCE in the project root.
# ------------------------------------------------------------------

module GslibIO

using FileIO
using Printf
using DelimitedFiles

using GeoStatsBase

# type aliases
const Array2or3{T} = Union{AbstractArray{T,2},AbstractArray{T,3}}

"""
    load(file)

Load grid properties from `file`.
"""
function load(file::File{format"GSLIB"})
  open(file) do f
    fs = stream(f)

    # skip header
    skipchars(_ -> false, fs, linecomment='#')

    # read dimensions
    dims = parse.(Int,     Tuple(split(readline(fs))))
    orig = parse.(Float64, Tuple(split(readline(fs))))
    spac = parse.(Float64, Tuple(split(readline(fs))))

    # read property names
    vars = Symbol.(split(readline(fs)))

    # read property values
    X = readdlm(fs)

    # create data dictionary
    data = (; zip(vars, eachcol(X))...)
    domain = RegularGrid(dims, orig, spac)

    georef(data, domain)
  end
end

"""
    load_legacy(filename, dims; origin=(0.,0.,0.), spacing=(1.,1.,1.), na=-999)

Load legacy GSLIB `filename` into a grid with `dims`, `origin` and `spacing`.
Optionally set the value used for missings `na`.
"""
function load_legacy(filename::AbstractString, dims::NTuple{3,Int};
                     origin=(0.,0.,0.), spacing=(1.,1.,1.), na=-999)
  open(filename) do fs
    # skip header
    readline(fs)

    # number of properties
    nvars = parse.(Int, readline(fs))

    # properties names
    vars = [Symbol(readline(fs)) for i in 1:nvars]

    # read property values
    X = readdlm(fs)

    # handle missing values
    replace!(X, na=>NaN)

    # create data dictionary
    data = (; zip(vars, eachcol(X))...)
    domain = RegularGrid(dims, origin, spacing)

    georef(data, domain)
  end
end

"""
    save(file, properties, dims; [optional parameters])

Save 1D `properties` to `file`, which originally had size `dims`.
"""
function save(file::File{format"GSLIB"},
              properties::AbstractVector, dims::Dims{N};
              origin=ntuple(i->0.0,N), spacing=ntuple(i->1.0,N),
              header="", propnames="") where {N}
  # default property names
  isempty(propnames) && (propnames = ["prop$i" for i=1:length(properties)])
  @assert length(propnames) == length(properties) "number of property names must match number of properties"

  # convert vector of names to a long string
  propnames = join(propnames, " ")

  # collect all properties in a big matrix
  P = reduce(hcat, properties)

  open(file, "w") do f
    # write header
    write(f, "# This file was generated with GslibIO.jl\n")
    !isempty(header) && write(f, "#\n# "*header*"\n")

    # write dimensions
    dimsstr = join([@sprintf "%i" i for i in dims], " ")
    origstr = join([@sprintf "%f" o for o in origin], " ")
    spacstr = join([@sprintf "%f" s for s in spacing], " ")
    write(f, dimsstr*"\n"*origstr*"\n"*spacstr*"\n")

    # write property name and values
    write(f, "$propnames\n")
    writedlm(stream(f), P, ' ')
  end
end

"""
    save(file, properties, [optional parameters])

Save 2D/3D `properties` by first flattening them into 1D properties.
"""
function save(file::File{format"GSLIB"}, properties::Vector{A};
              kwargs...) where {T,A<:Array2or3{T}}
  # sanity checks
  @assert length(Set(size.(properties))) == 1 "properties must have the same size"

  # retrieve grid size
  dims = size(properties[1])

  # flatten and proceed with pipeline
  flatprops = [vec(prop) for prop in properties]

  save(file, flatprops, dims, kwargs...)
end

"""
    save(file, property)

Save single 2D/3D `property` by wrapping it into a singleton collection.
"""
function save(file::File{format"GSLIB"},
              property::A; kwargs...) where {T,A<:Array2or3{T}}
  save(file, [property]; kwargs...)
end

"""
    save(file, sdata)

Save spatial data `sdata` with `RegularGrid` domain to `file`.
"""
function save(file::File{format"GSLIB"}, sdata::AbstractData)
  vars  = name.(variables(sdata))
  grid  = domain(sdata)
  table = values(sdata)
  save(file, collect(eachcol(table)), size(grid),
       origin=origin(grid), spacing=spacing(grid),
       propnames=vars)
end

end
